/*
 * modified for lecture by Fabian Fuchs (KIT, Karlsruhe) based on:
 * 
 Copyright (c) 2007, Distributed Computing Group (DCG)
                    ETH Zurich
                    Switzerland
                    dcg.ethz.ch

 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 - Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

 - Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the
   distribution.

 - Neither the name 'Sinalgo' nor the names of its contributors may be
   used to endorse or promote products derived from this software
   without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package projects.uebung3solution.nodes.nodeImplementations;


import java.awt.Color;
import java.awt.Graphics;
import java.util.LinkedList;

import javax.swing.JOptionPane;

import projects.defaultProject.nodes.timers.MessageTimer;
import projects.uebung2.nodes.messages.DefaultMessage;
import sinalgo.configuration.Configuration;
import sinalgo.configuration.WrongConfigurationException;
import sinalgo.gui.helper.NodeSelectionHandler;
import sinalgo.gui.transformation.PositionTransformation;
import sinalgo.nodes.Node;
import sinalgo.nodes.Position;
import sinalgo.nodes.edges.Edge;
import sinalgo.nodes.messages.Inbox;
import sinalgo.nodes.messages.Message;
import sinalgo.runtime.GUIRuntime;
import sinalgo.runtime.Main;
import sinalgo.tools.Tools;
import sinalgo.tools.Tuple;




/**
 * A leader election node 
 */
public class GreedyFacetRoutingNode extends Node {

	/* store the current round */
	private int round;

	/* add additional variables / flags here */
	
	/* available states */
	public static final int GREEDY = 0;
	public static final int FACET = 1;
	
	/* facet routing information */
	public class FRData {
		public int bestFRNodeID;
		public double bestFRDistance;
		public Position lastPosition;
		public FRData() {
			bestFRNodeID = -1;
			bestFRDistance = 2*Configuration.dimX+Configuration.dimY+Configuration.dimZ;
		}
	}
	
	/* payload for message */
	public class GRFPayload<T> {
		public Position toPosition;
		public T payload;
		public int state;
		public FRData frData;
		public GRFPayload (Position to, T pl) {
			toPosition = to;
			payload = pl;
			state = GREEDY;
			frData = new FRData();
		}
	}
	
	/* 
	 * finds the closest neighbor based on this.neighborPositions
	 * and return the ID of the closest neighbor (or the own ID)
	 */
	private int getNextNodeID(Position destination) {
		
		//init to the node itself
		double minDist = this.computeDistance(this.getPosition(), destination);
		int routeTo = this.ID;
		double tmp = 0;
		//try to find a neighbor with better values
		for (Edge e : this.outgoingConnections) {
			tmp = this.computeDistance(e.endNode.getPosition(), destination);
			if (tmp < minDist) {
				minDist = tmp;
				routeTo = e.endNode.ID;
			}
		}
		//return best value
		return routeTo;
	}
	
	/* 
	 * by: http://stackoverflow.com/questions/9970281/java-calculating-the-angle-between-two-points-in-degrees
	 * works only for two dimensions!
	 */
	private float getAngle(Position from, Position to) {
	    float angle = (float) Math.toDegrees(Math.atan2(to.yCoord - from.yCoord, to.xCoord - from.xCoord));
	    if(angle < 0){
	        angle += 360;
	    }
	    return angle;
	}
	
	/* 
	 * packet got stuck at this node, compute the first node of the
	 * facet that should be searched
	 */
	private int getFirstFacetNodeID(Position destination) {
		return getNextFacetNodeID(destination);
	}
	
	/*
	 * received packet in FACET state, compute next node of the facet
	 */
	private int getNextFacetNodeID(Position fromPos) {
		double inAngle = getAngle(this.getPosition(),fromPos);
		double tmp=0,best=720;
		int bestID = this.ID;
		//Tools.appendToOutput("ID " + this.ID + " inAngle " + inAngle + "\n");
		for (Edge e : this.outgoingConnections) {
			tmp = getAngle(this.getPosition(),e.endNode.getPosition());
			if (tmp < inAngle) tmp += 360; //remove case that tmp is smaller than inAngle
			//Tools.appendToOutput("ID " + this.ID + " tmpAngle " + tmp + "\n");
			if (tmp - inAngle < best && tmp != inAngle) {
				best = tmp - inAngle;
				//Tools.appendToOutput("" + best + "\n");
				bestID = e.endNode.ID;
			}
		}
		//Tools.appendToOutput("ID " + this.ID + " best Angle " + best + "\n");
		if (bestID == this.ID) {
			Tools.appendToOutput("Isolated node?");
		}
		return bestID;
	}
	
	private double computeDistance(Position from, Position to) {
		double xd = Math.abs(from.xCoord - to.xCoord);
		double yd = Math.abs(from.yCoord - to.yCoord);
		double zd = Math.abs(from.zCoord - to.zCoord);
		return Math.sqrt(xd*xd+yd*yd+zd*zd);
	}
	
	@Override
	public void handleMessages(Inbox inbox) {
		while(inbox.hasNext()) {
			Message msg = inbox.next();
			if(msg instanceof DefaultMessage<?>) {
				if (((DefaultMessage<?>) msg).getLoad() instanceof GRFPayload<?>) {
					//unpack packet
					GRFPayload<?> packet = (GRFPayload<?>)((DefaultMessage<?>) msg).getLoad();
					//I assume a String payload
					if (!(packet.payload instanceof String)){
						Tools.appendToOutput("Payload error. Payload: " + packet.payload);
					}
					//is packet for me?
					if (this.getPosition() == packet.toPosition) {
						Tools.appendToOutput("ID " + this.ID + ": Received message with payload " + (String)(packet.payload) + "\n");
					} else {
						//FACET or GREEDY routing?
						if (packet.state == FACET) { //if in FR state: 
							if (packet.frData.bestFRNodeID == this.ID) {
								//restart Greedy
								packet.state = GREEDY;
							} else {
								int routeTo = this.getNextFacetNodeID(packet.frData.lastPosition); //position of neighbor that sent the packet!
								Tools.appendToOutput("ID " + this.ID + ": Facet routing to " + routeTo + "\n");
								double tmpDist = this.computeDistance(this.getPosition(), packet.toPosition);
								if (tmpDist < packet.frData.bestFRDistance) {
									Tools.appendToOutput("ID " + this.ID + ": new best node on facet!\n");
									packet.frData.bestFRDistance = tmpDist; 
									packet.frData.bestFRNodeID = this.ID;
								}
								packet.frData.lastPosition = this.getPosition();
								this.send(msg, Tools.getNodeByID(routeTo));
							}
						}
						//do not use else if, in case facet routing ends
						if (packet.state == GREEDY) { //if in GR state: 					
							//find closest neighbor
							int routeTo = getNextNodeID(packet.toPosition);
							//if closest neighbor is not the node itself, transmit the packet to the neighbor
							if (routeTo != this.ID) {
								Tools.appendToOutput("ID " + this.ID + ": Forward message\n");
								this.send(msg, Tools.getNodeByID(routeTo));
							} else {
								Tools.appendToOutput("ID " + this.ID + ": Message got stuck\n");
								packet.state = FACET;
								routeTo = getFirstFacetNodeID(packet.toPosition);
								Tools.appendToOutput("ID " + this.ID + ": Start Facet routing to " + routeTo + "\n");
								packet.frData.bestFRDistance = this.computeDistance(this.getPosition(), packet.toPosition);
								packet.frData.bestFRNodeID = this.ID;
								packet.frData.lastPosition = this.getPosition();
								this.send(msg, Tools.getNodeByID(routeTo));
							}
						} else if (packet.state == FACET) {
						}
					} 
				}
			}
		}
	}

	@Override
	public void preStep() {
		//executed before each 'step' (i.e. timers and handleMessages)
		//I would personally recommend to only use the postStep method for simplicity
	}
	
	@Override
	public void postStep() {
		//executed after each 'step' (i.e. timers and handleMessages)
		round++;
	}

	@Override
	public void init() {
		// initialize the node / read configuration values
		round = 0;
	}

	@Override
	public void neighborhoodChange() {

	}
		
	/* (non-Javadoc)
	 * @see sinalgo.nodes.Node#draw(java.awt.Graphics, sinalgo.gui.transformation.PositionTransformation, boolean)
	 */
	public void draw(Graphics g, PositionTransformation pt, boolean highlight) {
		//set the color of this node, this allows us to see how the most important maxIDs spread
		
		this.setColor(Color.BLACK); 
		String text = Integer.toString(this.ID);
		// draw the node as a circle with the text inside
		super.drawNodeAsDiskWithText(g, pt, highlight, text, 10, Color.YELLOW);
	}


	@Override
	public String toString() {
		return "Node " + this.ID; 
	}

	@Override
	public void checkRequirements() throws WrongConfigurationException {
		//does configuration fit the nodes requirements? 
		
	}
	
	/*
	 * Starts the greedy routing of a message
	 */
	@NodePopupMethod(menuText="StartRouting")
	public void startRouting() {
		
		class NodeSelectionHandlerGreedyRouting implements NodeSelectionHandler {
			public GreedyFacetRoutingNode sender;
			public void handleNodeSelectedEvent(Node n) { //n = null implies abort by user
				if (n == null) {
					Tools.appendToOutput("GreedyFacetRouting: Abort by user\n");
					return;
				}
				sender.doStartRouting(n);
			}
			NodeSelectionHandlerGreedyRouting(GreedyFacetRoutingNode sender) {
				this.sender = sender;
			}

		}
		NodeSelectionHandlerGreedyRouting nsh = new NodeSelectionHandlerGreedyRouting(this);

		Tools.getNodeSelectedByUser(nsh, "Select the destination node.");
	}
	
	public void doStartRouting(Node to) {
		//create payload
		String payload = JOptionPane.showInputDialog(null, "Enter a payload String");

		Position destination = to.getPosition();
		//compute the best neighbor 
		int routeTo = getNextNodeID(destination);
		//if closest neighbor is not the node itself, transmit the packet to the neighbor
		if (routeTo != this.ID) {
			DefaultMessage<GRFPayload<String>> msg = new DefaultMessage<GRFPayload<String>>(new GRFPayload<String>(destination, payload));
			MessageTimer msgTimer = new MessageTimer(msg, Tools.getNodeByID(routeTo));
			msgTimer.startRelative(1, this);
			Tools.appendToOutput("ID: " + this.ID + " Start Greedy Routing a packet to " + to.ID + " with position " + destination + "\n");		
		} else {
			GRFPayload<String> packet = new GRFPayload<String>(destination, payload);
			packet.state = FACET;
			routeTo = getFirstFacetNodeID(packet.toPosition);
			Tools.appendToOutput("ID " + this.ID + ": Start Facet routing to " + routeTo + "\n");
			packet.frData.bestFRDistance = this.computeDistance(this.getPosition(), packet.toPosition);
			packet.frData.bestFRNodeID = this.ID;
			packet.frData.lastPosition = this.getPosition();
			DefaultMessage<GRFPayload<String>> msg = new DefaultMessage<GRFPayload<String>>(packet);
			MessageTimer msgTimer = new MessageTimer(msg, Tools.getNodeByID(routeTo));
			msgTimer.startRelative(1, this);
		}
	}
}
